package	jp.co.sds_corp.util.jmx.rmi.client;

import java.io.ObjectInputStream;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Set;
import java.util.Vector;

import javax.management.Attribute;
import javax.management.AttributeList;
import javax.management.AttributeNotFoundException;
import javax.management.InstanceAlreadyExistsException;
import javax.management.InstanceNotFoundException;
import javax.management.IntrospectionException;
import javax.management.InvalidAttributeValueException;
import javax.management.ListenerNotFoundException;
import javax.management.MBeanException;
import javax.management.MBeanInfo;
import javax.management.MBeanRegistrationException;
import javax.management.NotCompliantMBeanException;
import javax.management.Notification;
import javax.management.NotificationFilter;
import javax.management.NotificationListener;
import javax.management.ObjectInstance;
import javax.management.ObjectName;
import javax.management.OperationsException;
import javax.management.QueryExp;
import javax.management.ReflectionException;
import javax.management.loading.ClassLoaderRepository;
import javax.naming.InitialContext;

import jp.co.sds_corp.util.jmx.rmi.JMXConnectorMBean;
import jp.co.sds_corp.util.jmx.rmi.RMIConnector;
import jp.co.sds_corp.util.jmx.rmi.RMINotificationListener;
import jp.co.sds_corp.util.jmx.rmi.server.ObjectHandler;

public class RMIClientConnectorImpl
	implements JMXConnectorMBean, RMIClientConnectorImplMBean
{
	private RMIConnector		mRemoteConnector;
	private Object				mServer = "";
	private Hashtable<Object,Object>	mHandbackPool = new Hashtable<Object,Object>();
	private Vector<Listener>	mListeners = new Vector<Listener>();

	public RMIClientConnectorImpl( String pServerName ) {
		super();
		start( pServerName );
	}

	public void start( Object pServer ) throws IllegalArgumentException {
		if( pServer == null ) {
			throw new IllegalArgumentException( "Server cannot be null. "
				+ "To close the connection use stop()" );
		}
		try {
			InitialContext lNamingContext = new InitialContext();
			System.out.println( "RMIClientConnectorImp.start(), got Naming Context: " +	lNamingContext +
				", environment: " + lNamingContext.getEnvironment() +
				", name in namespace: " + lNamingContext.getNameInNamespace()
			);
			mRemoteConnector = (RMIConnector) new InitialContext().lookup( "jp.co.sds_corp.util.jmx:" + pServer + ":rmi" );
			System.err.println( "RMIClientConnectorImpl.start(), got remote connector: " + mRemoteConnector );
			mServer = pServer;
		} catch( Exception ex ) {
			ex.printStackTrace();
		}
	}

	public void stop() {
		System.out.println( "RMIClientConnectorImpl.stop(), start" );
		Iterator i = mListeners.iterator();
		while( i.hasNext() ) {
			try {
				Listener lRemoteListener = (Listener) i.next();
				System.out.println( "RMIClientConnectorImpl.stop(), remove listener: " +
					lRemoteListener );
				try {
					mRemoteConnector.removeNotificationListener(
						lRemoteListener.getObjectName(),
						lRemoteListener );
				} catch( RemoteException re ) {
					re.printStackTrace();
				} finally {
					i.remove();
				}
			} catch( Exception ex ) {
				ex.printStackTrace();
			}
		}
		mRemoteConnector = null;
		mServer = "";
	}
	
	public boolean isAlive() {
		return mRemoteConnector != null;
	}

	public String getServerDescription() {
		return "" + mServer;
	}

	public Object instantiate( String pClassName )
		throws ReflectionException, MBeanException
	{
		try {
			return mRemoteConnector.instantiate( pClassName );
		} catch( RemoteException re ) {
			return null;
		}
	}

	public Object instantiate( String pClassName, ObjectName pLoaderName )
		throws ReflectionException, MBeanException, InstanceNotFoundException
	{
		try {
			return mRemoteConnector.instantiate( pClassName, pLoaderName );
		} catch( RemoteException re ) {
			return null;
		}
	}

	public Object instantiate( String pClassName, Object[] pParams,
								String[] pSignature )
		throws ReflectionException, MBeanException
	{
		try {
			return mRemoteConnector.instantiate( pClassName, pParams, pSignature );
		} catch( RemoteException re ) {
			return null;
		}
	}

	public Object instantiate( String pClassName, ObjectName pLoaderName,
								Object[] pParams, String[] pSignature )
		throws ReflectionException, MBeanException, InstanceNotFoundException
	{
		try {
			return mRemoteConnector.instantiate( pClassName, pLoaderName, pParams, pSignature );
		} catch( RemoteException re ) {
			return null;
		}
	}

	public ObjectInstance createMBean( String pClassName, ObjectName pName )
		throws ReflectionException, InstanceAlreadyExistsException,
				MBeanRegistrationException, MBeanException,
				NotCompliantMBeanException
	{
		try {
			return mRemoteConnector.createMBean( pClassName, pName );
		} catch( RemoteException re ) {
			//AS Not a good style but for now
			return null;
		}
	}

	public ObjectInstance createMBean( String pClassName, ObjectName pName,
							ObjectName pLoaderName )
		throws ReflectionException, InstanceAlreadyExistsException,
				MBeanRegistrationException, MBeanException,
				NotCompliantMBeanException, InstanceNotFoundException
	{
		try {
			return mRemoteConnector.createMBean( pClassName, pName, pLoaderName );
		} catch( RemoteException re ) {
			return null;
		}
	}

	public ObjectInstance createMBean( String pClassName, ObjectName pName,
							Object[] pParams, String[] pSignature )
		throws ReflectionException, InstanceAlreadyExistsException,
				MBeanRegistrationException, MBeanException,
				NotCompliantMBeanException
	{
		try {
			return mRemoteConnector.createMBean( pClassName, pName, pParams, pSignature );
		} catch( RemoteException re ) {
			return null;
		}
	}

	public ObjectInstance createMBean( String pClassName, ObjectName pName,
			ObjectName pLoaderName, Object[] pParams, String[] pSignature )
		throws ReflectionException, InstanceAlreadyExistsException,
				MBeanRegistrationException, MBeanException,
				NotCompliantMBeanException, InstanceNotFoundException
	{
		try {
			return mRemoteConnector.createMBean( pClassName, pName, pLoaderName, pParams, pSignature );
		} catch( RemoteException re ) {
			return null;
		}
	}

	public ObjectInstance registerMBean( Object pObject, ObjectName pName )
		throws InstanceAlreadyExistsException, MBeanRegistrationException,
				NotCompliantMBeanException
	{
		try {
			return mRemoteConnector.registerMBean( pObject, pName );
		} catch( RemoteException re ) {
			return null;
		}
	}

	public void unregisterMBean( ObjectName pName )
		throws InstanceNotFoundException, MBeanRegistrationException
	{
		try {
			mRemoteConnector.unregisterMBean( pName );
		} catch( RemoteException re ) {
		}
	}

	public ObjectInstance getObjectInstance( ObjectName pName )
		throws InstanceNotFoundException
	{
		try {
			return mRemoteConnector.getObjectInstance( pName );
		} catch( RemoteException re ) {
			return null;
		}
	}

	public Set<ObjectInstance> queryMBeans( ObjectName pName, QueryExp pQuery ) {
		try {
			return mRemoteConnector.queryMBeans( pName, pQuery );
		} catch( RemoteException re ) {
			re.printStackTrace();
			return null;
		}
	}

	public Set<ObjectName> queryNames( ObjectName pName, QueryExp pQuery ) {
		try {
			return mRemoteConnector.queryNames( pName, pQuery );
		} catch( RemoteException re ) {
			return null;
		}
	}

	public boolean isRegistered( ObjectName pName ) {
		try {
			return mRemoteConnector.isRegistered( pName );
		} catch( RemoteException re ) {
			return false;
		}
	}

	public boolean isInstanceOf( ObjectName pName, String pClassName )
		throws InstanceNotFoundException
	{
		try {
			return mRemoteConnector.isInstanceOf( pName, pClassName );
		} catch( RemoteException re ) {
			return false;
		}
	}

	public Integer getMBeanCount( ) {
		try {
			return mRemoteConnector.getMBeanCount();
		} catch( RemoteException re ) {
			return null;
		}
	}

	public Object getAttribute( ObjectName pName, String pAttribute )
		throws MBeanException, AttributeNotFoundException,
				InstanceNotFoundException, ReflectionException
	{
		try {
			return mRemoteConnector.getAttribute( pName, pAttribute );
		} catch( RemoteException re ) {
			return null;
		}
	}

	public AttributeList getAttributes( ObjectName pName, String[] pAttributes )
		throws InstanceNotFoundException, ReflectionException
	{
		try {
			return mRemoteConnector.getAttributes( pName, pAttributes );
		} catch( RemoteException re ) {
			return null;
		}
	}

	public void setAttribute( ObjectName pName, Attribute pAttribute )
		throws InstanceNotFoundException, AttributeNotFoundException,
				InvalidAttributeValueException, MBeanException,
				ReflectionException
	{
		try {
			mRemoteConnector.setAttribute( pName, pAttribute );
		} catch( RemoteException re ) {
		}
	}

	public AttributeList setAttributes( ObjectName pName,
								AttributeList pAttributes )
		throws InstanceNotFoundException, ReflectionException
	{
		try {
			return mRemoteConnector.setAttributes( pName, pAttributes );
		} catch( RemoteException re ) {
			return null;
		}
	}

	public Object invoke( ObjectName pName, String pActionName,
								Object[] pParams, String[] pSignature )
		throws InstanceNotFoundException, MBeanException, ReflectionException
	{
		try {
			return mRemoteConnector.invoke( pName, pActionName, pParams, pSignature );
		} catch( RemoteException re ) {
			 re.printStackTrace();
			return null;
		}
	}

	public String getDefaultDomain() {
		try {
			return mRemoteConnector.getDefaultDomain();
		} catch( RemoteException re ) {
			return null;
		}
	}

	public void addNotificationListener( ObjectName pName,
					NotificationListener pListener, NotificationFilter pFilter,
					Object pHandback )
		throws InstanceNotFoundException
	{
		try {
			Listener lRemoteListener = new Listener(
				pListener, pHandback, pName );
			UnicastRemoteObject.exportObject( lRemoteListener );
			mRemoteConnector.addNotificationListener(
				pName, lRemoteListener, pFilter, lRemoteListener.getHandback());
			mListeners.addElement( lRemoteListener );
		} catch( RemoteException re ) {
			re.printStackTrace();
		}
	}

	public void addNotificationListener( ObjectName pName, ObjectName pListener,
		NotificationFilter pFilter, Object pHandback )
		throws InstanceNotFoundException, UnsupportedOperationException
	{
		throw new UnsupportedOperationException(
			"A connector cannot support this method" );
	}

	public void removeNotificationListener( ObjectName pName,
							ObjectName pListener,
							NotificationFilter filter, Object handback )
		throws InstanceNotFoundException, ListenerNotFoundException
	{
		throw new UnsupportedOperationException(
			"A connector cannot support this method" );
	}
	public void removeNotificationListener( ObjectName pName,
							NotificationListener pListener,
							NotificationFilter filter, Object handback )
		throws InstanceNotFoundException, ListenerNotFoundException
	{
		throw new UnsupportedOperationException(
			"A connector cannot support this method" );
	}

	public void removeNotificationListener( ObjectName pName,
									NotificationListener pListener )
		throws InstanceNotFoundException, ListenerNotFoundException
	{
		Iterator i = mListeners.iterator();
		while( i.hasNext() ) {
			Listener lListener = (Listener) i.next();
			if( new Listener( pListener, null, pName ).equals( lListener ) ) {
				try {
					mRemoteConnector.removeNotificationListener( pName, lListener );
				} catch( RemoteException re ) {
					re.printStackTrace();
				} finally {
					i.remove();
				}
			}
		}
	}

	public void removeNotificationListener( ObjectName pName, ObjectName pListener )
		throws InstanceNotFoundException, ListenerNotFoundException,
				UnsupportedOperationException
	{
		throw new UnsupportedOperationException( "A connector cannot support this method" );
	}

	public MBeanInfo getMBeanInfo( ObjectName pName )
		throws InstanceNotFoundException, IntrospectionException,
				ReflectionException
	{
		try {
			return mRemoteConnector.getMBeanInfo( pName );
		} catch( RemoteException re ) {
			return null;
		}
	}

	/**
	* @deprecated
	*/
	public ObjectInputStream deserialize( ObjectName pName, byte[] pData )
		throws InstanceNotFoundException, OperationsException,
				UnsupportedOperationException
	{
		throw new UnsupportedOperationException( "Remotely this method cannot be supported" );
	}

	/**
	* @deprecated
	*/
	public ObjectInputStream deserialize( String pClassName, byte[] pData )
		throws OperationsException, ReflectionException,
				UnsupportedOperationException
	{
		throw new UnsupportedOperationException( "Remotely this method cannot be supported" );
	}

	/**
	* @deprecated
	*/
	public ObjectInputStream deserialize( String pClassName,
								ObjectName pLoaderName, byte[] pData )
		throws InstanceNotFoundException, OperationsException,
				ReflectionException, UnsupportedOperationException
	{
		throw new UnsupportedOperationException( "Remotely this method cannot be supported" );
	}

	public ClassLoaderRepository getClassLoaderRepository() {
		throw new UnsupportedOperationException( "Remotely this method cannot be supported" );
	}
	public ClassLoader getClassLoader( ObjectName name ) {
		throw new UnsupportedOperationException( "Remotely this method cannot be supported" );
	}
	public ClassLoader getClassLoaderFor( ObjectName name ) {
		throw new UnsupportedOperationException( "Remotely this method cannot be supported" );
	}
	public String[] getDomains() {
		throw new UnsupportedOperationException( "Remotely this method cannot be supported" );
	}

	public class Listener implements RMINotificationListener {
		private NotificationListener			mLocalListener;
		private ObjectHandler					mHandbackHandler;
		private Object							mHandback;
		private ObjectName						mObjectName;
		
		public Listener( NotificationListener pLocalListener,
			Object pHandback, ObjectName pName )
		{
			mLocalListener = pLocalListener;
			mHandback = pHandback;
			mHandbackHandler = new ObjectHandler( this.toString() );
			mObjectName = pName;
		}

		public void handleNotification( Notification pNotification,
						Object pHandback )
			throws RemoteException
		{
			Object lHandback;
			if( mHandbackHandler.equals( pHandback ) ) {
				lHandback = mHandback;
			} else {
				lHandback = pHandback;
			}
			mLocalListener.handleNotification( pNotification, lHandback );
		}
		
		public ObjectHandler getHandback() {
			return mHandbackHandler;
		}
		public ObjectName getObjectName() {
			return mObjectName;
		}
		public NotificationListener getLocalListener() {
			return mLocalListener;
		}
		public boolean equals( Object pTest ) {
			if( pTest instanceof Listener ) {
				return mLocalListener.equals(
					( (Listener) pTest).mLocalListener
				);
			}
			return false;
		}
		public int hashCode() {
			return mLocalListener.hashCode();
		}
	}
}	

